import rclpy
from rclpy.node import Node
from nav_msgs.msg import OccupancyGrid
from geometry_msgs.msg import PoseStamped
from nav2_msgs.action import NavigateToPose
from rclpy.action import ActionClient
import random
import time

class RandomNavigator(Node):
    def __init__(self):
        super().__init__('random_navigator')

        # Subscribers and publishers
        self.map_subscription = self.create_subscription(
            OccupancyGrid,
            '/map',
            self.map_callback,
            10
        )

        self.goal_publisher = self.create_publisher(PoseStamped, '/goal_pose', 10)

        # Action client for navigation
        self.navigation_client = ActionClient(self, NavigateToPose, 'navigate_to_pose')

        # Internal state
        self.map_data = None
        self.get_logger().info("Random Navigator initialized. Waiting for map...")

    def map_callback(self, msg):
        self.map_data = msg
        self.get_logger().info("Map received.")

    def get_random_goal(self):
        if not self.map_data:
            self.get_logger().warn("Map data not available.")
            return None

        resolution = self.map_data.info.resolution
        origin_x = self.map_data.info.origin.position.x
        origin_y = self.map_data.info.origin.position.y

        free_indices = [
            idx for idx, value in enumerate(self.map_data.data) if value == 0
        ]

        if not free_indices:
            self.get_logger().warn("No free space available in the map.")
            return None

        random_index = random.choice(free_indices)
        grid_x = random_index % self.map_data.info.width
        grid_y = random_index // self.map_data.info.width

        goal_x = origin_x + grid_x * resolution
        goal_y = origin_y + grid_y * resolution

        return goal_x, goal_y

    def send_goal(self, goal_x, goal_y):
        goal_msg = PoseStamped()
        goal_msg.header.frame_id = 'map'
        goal_msg.header.stamp = self.get_clock().now().to_msg()
        goal_msg.pose.position.x = goal_x
        goal_msg.pose.position.y = goal_y
        goal_msg.pose.orientation.w = 1.0

        self.get_logger().info(f"Publishing goal: x={goal_x}, y={goal_y}")
        self.goal_publisher.publish(goal_msg)

        # Use NavigateToPose action to monitor navigation
        goal_action_msg = NavigateToPose.Goal()
        goal_action_msg.pose = goal_msg
        self.navigation_client.wait_for_server()
        self.navigation_client.send_goal_async(goal_action_msg, feedback_callback=self.feedback_callback)

    def feedback_callback(self, feedback):
        self.get_logger().info(f"Navigation feedback: {feedback.feedback}")

    def monitor_navigation(self, timeout=30):
        start_time = time.time()
        while time.time() - start_time < timeout:
            rclpy.spin_once(self, timeout_sec=1.0)
            if self.navigation_completed():
                return True

        self.get_logger().warn("Navigation timeout!")
        return False

    def navigation_completed(self):
        # Placeholder logic for checking completion; integrate with result feedback
        return False

    def run(self):
        while rclpy.ok():
            goal = self.get_random_goal()
            if goal:
                self.send_goal(goal[0], goal[1])
                success = self.monitor_navigation(timeout=30)
                if not success:
                    self.get_logger().warn("Retrying with a new goal.")
            else:
                self.get_logger().warn("Failed to generate a valid goal.")
            
            time.sleep(1)

def main():
    rclpy.init()
    navigator = RandomNavigator()
    try:
        navigator.run()
    except KeyboardInterrupt:
        navigator.get_logger().info("Shutting down Random Navigator.")
    finally:
        navigator.destroy_node()
        rclpy.shutdown()

